#' Create a skeleton for a new source package
#'
#' @param packageName name of the package (and).
#' @param build.dir   name of directory to dump built package in (default
#' 'bin').
#' @param hg          boolean, whether to set this up as a mercurial
#' project (initialises the project in mercurial)
#' @param hg.commit   whether to do an initial commit of the mercurial repo
#' @param empty       boolean, whether to create an entirely empty package
#' (ie ignores list and env arguments).
#' @param path        path that the entire project directory will be *placed in*.
#' @param version     starting version of package, string ('0.1'): put into
#' DESCRIPTION file.
#' @param force       boolean, whether to *REMOVE AND WRITE OVER* existing
#'                    directories
#' @param ... passed to \code{\link[utils]{package.skeleton}}
#' @details
#' The difference between this and the base
#' \code{\link[utils]{package.skeleton}} is that this sets up the project
#' like so:
#'
#' \preformatted{
#' path
#' |- packageName-package     The "project" directory
#'    |- Makefile     A makefile for compiling various bits
#'    |- README       A readme for the developer
#'    |- build.dir    The output directory things are built in
#'    |- packageName  The actual R package directory
#'       |- R  
#'       |- inst      ... etc (packages created by R's built-in
#'       |- man                'package.skeleton' function).
#' }
#'
#' The in-built \code{\link[utils]{package.skeleton}} function only creates
#' the internal 'packageName' directory.
#'
#' This version sets up the package in an environment more suited to a
#' developer, with a handy Makefile to generate documentation ('make doc'),
#' generate the package ('make' or 'make package'), check the package
#' ('make check'), and/or install it ('make install').
#'
#' It also separates source code (in the packageName dir) from the build
#' and check outputs (in the build.dir dir).
#'
#' If the mercurial versioning system is installed on the system, it can
#' optionally create a mercurial repository for the project, with a
#' pre-generated \code{.hgignore} file which ignores generated files like
#' everything in build.dir, Rd files, etc.
#'
#' The Makefile contains the targets 'doc' (to build documentation using
#' the roxygen2 package), 'package' (to build the package using \code{R CMD
#' BUILD}), 'check' (to check the package with \code{R CMD CHECK}), and
#' 'install' (to install the package with \code{R CMD INSTALL}).
#'
#' @examples \dontrun{
#' my.package.skeleton('newPackage',hg=TRUE)
#' }
#' @export
my.package.skeleton <- function(packageName, path=".", build.dir='bin', 
                                hg=FALSE, hg.commit=hg, force=FALSE, 
                                empty=FALSE, version='0.1', ...)
{
    safe.dir.create <- function(path) {
        dirTest <- function(x) !is.na(isdir <- file.info(x)$isdir) & 
            isdir
            if (!dirTest(path) && !dir.create(path)) 
                stop(gettextf("cannot create directory '%s'",path),
                                     domain = NA)
    }

    dir <- file.path(path,sprintf('%s-package',packageName))
    if ( file.exists(dir) )
        if ( force )
            unlink(dir,recursive=TRUE)
        else
            stop(sprintf("Directory '%s' already exists",dir))

    message("Creating project directory...")
    safe.dir.create(dir)
tryCatch({
    message("Creating build directory...")
    # make build dir
    safe.dir.create(file.path(dir,build.dir))

    # make package in there
    # NOTE: there's a bug if you have nothing in your
    #  environment when you create the package:
    # it will say "!have: invalid argument type".
    # Making something (that doesn't get included anyway) is the
    # workaround.
    DELETE_LATER <- FALSE
    if ( !is.na(i<-match('env', names(list(...)))) )
        env <- list(...)[[i]]
    else
        if ( empty )
            env <- new.env()
        else
            env <- globalenv()

    if ( (DELETE_LATER <- (length(ls(env))==0) ) ) 
        assign('.foo',1,envir=env)
    message("Creating R package directory...")
    suppressMessages(package.skeleton(packageName,path=dir,environment=env,...))

    # put in version number:
    DESC <- readLines(file.path(dir,packageName,'DESCRIPTION'))
    DESC[grep('^Version: ',DESC)] <- sprintf('Version: %s',trim(version))
    writeLines(DESC,file.path(dir,packageName,'DESCRIPTION'))

    if ( DELETE_LATER ) {
        # there was nothing there anyway since the environment was blank,
        # so it's safe to remove everything:
        file.remove(list.files(file.path(dir,packageName,'R'),full.names=TRUE),
                    list.files(file.path(dir,packageName,'man'),full.names=TRUE))
        rm('.foo',envir=env)
    }
    # just in case
    safe.dir.create(file.path(dir,packageName,'R'))
    safe.dir.create(file.path(dir,packageName,'man'))

    # kill the "read-and-delete-me"
    file.remove( file.path(dir,packageName,"Read-and-delete-me") )

    message("Creating Makefile...")
    # Makefile
    cat(sprintf(
'default_target: all
#=============================================================================
RCMD:=$(shell /usr/bin/which R)
RSCRIPT:=$(shell /usr/bin/which Rscript)
BUILD_DIR=%s
PACKAGE_NAME=%s
PACKAGE_DIR=$(PACKAGE_NAME)
VSN:=(shell grep -oP \'(?<=^Version: )(\\d+\\.\\d+(?:\\.\\d+)?) *$$\' $(PACKAGE_DIR)/DESCRIPTION)
#=============================================================================
# Set environment variables for the build.
all: clean package
doc: 
	@echo "Making documentation..."
	"$(RSCRIPT)" -e "library(utils); library(methods); library(roxygen2); roxygenise(\'$(PACKAGE_DIR)\')";
# remove dir inst if empty (roxygen bug)
	@rmdir --i $(PACKAGE_DIR)/inst;

package: doc
	@if [ ! -d "$(BUILD_DIR)" ]; then \\
		mkdir $(BUILD_DIR); \\
	fi
	(cd $(BUILD_DIR); "$(RCMD)" CMD build ../$(PACKAGE_DIR);)

check: package
	"$(RCMD)" CMD check -o $(BUILD_DIR) $(PACKAGE_DIR);

install: package
	@(if [ "${R_LIBS_USER}" != "" ] ; then \\
		echo "** Installing library" ;\\
		"$(RCMD)" CMD REMOVE -l "${R_LIBS_USER}" $(PACKAGE_NAME) ;\\
		"$(RCMD)" CMD INSTALL -l "${R_LIBS_USER}" $(BUILD_DIR)/$(PACKAGE_NAME)_$(VSN).tar.gz ;\\
	else \\
		echo "** Installing library as root" ;\\
		"$(RCMD)" CMD REMOVE $(PACKAGE_NAME) ;\\
		"$(RCMD)" CMD INSTALL $(BUILD_DIR)/$(PACKAGE_NAME)_$(VSN).tar.gz ;\\
	fi)
dev-install: clean check install
zip: package
	"$(RCMD)" CMD INSTALL -l $(BUILD_DIR) $(BUILD_DIR)/$(PACKAGE_NAME)_$(VSN).tar.gz
	(cd $(BUILD_DIR); zip -r $(PACKAGE_NAME)_$(VSN).zip $(PACKAGE_NAME); rm -rf $(PACKAGE_NAME);)

clean:
	rm -f $(PACKAGE_DIR)/man/*.Rd;
	rm -rf $(BUILD_DIR)/*;
.PHONY: clean
', build.dir, packageName), file=file.path(dir,'Makefile'))

    # Readme.
    message("Creating Readme...")
cat(sprintf('
FOR THE USER 
============
Get a zip/tar file with the package already generated,
 instead of installing from source.
> install.packages(\"%s_1.0.tar.gz\",repos=NULL)

FOR THE DEVELOPER
=================
R files are in %s/R.
Roxygen is used to generate documentation: all Rd files are generated and get written over.
A Makefile is provided for building.

From the terminal/command line [need to be able to run "R CMD ...."]
 make
 make install

This essentially runs (from directory %s):
 (run in R to generate docs):
 library(roxygen2)
 roxygenise("../%s")
 (run in shell):
 R CMD roxygen -d %s
 cd bin
 R CMD build ../%s
 R CMD INSTALL %s*.tar.gz
 cd ..

If you\'re developing and want to check the package (ie R CMD check), do
 make check

To make a windows .zip file, do:
 make zip

NOTE: if you\'ve set up a mercurial repository, there is a file
.precommit.sh that increments version numbers on every commit -- to disable
this feature edit .hg/hgrc and remove the "pre-commit = ../.precommit.sh"
line.

FROM WITHIN R
=============
This may or may not work...

Go to R, and set working dir to this directory.
> install.packages("%s",repos=NULL,type="source")

There is a possibility the documentation will be out of sync with the R files
 as I regenerate them afresh every time I modify the package -- if you want to
 be sure that the documentation is fresh, then:

> library(roxygen2)
> roxygenise("%s")
> install.packages("%s",repos=NULL,type="source")

Then, if all goes well:
> library(%s)
', packageName, packageName, build.dir, packageName, packageName,
packageName, packageName, packageName, packageName, packageName,
packageName),file=file.path(dir,'README'))

    # set up hg
    if ( hg ) {
        message("Setting up mercurial repository...")
        # .hgignore
        cat('
syntax:glob

*.roxygen
*.Rcheck
*.Rd

bin/*', file=file.path(dir,'.hgignore'))

        # hg init
        HG <- getProgram(c('hg','hg.exe'))
        if ( is.null(HG) ) {
            message(sprintf(paste(
              "Couldn't find 'hg' executable in your path,",
              "go to '%s' and type 'hg init'",
              "to initialise the repository.",collapse=" "),
                            file.path(dir)))
        } else {
            wd <- getwd()
            setwd(dir)
            # init
            system(sprintf('%s init',HG[1]),intern=TRUE)
            # precommit (automatic versioning)
cat(sprintf('#!/bin/sh
# Get version number: 
PACKAGE_DIR=%s

VSN=$(grep -oP \'(?<=^Version: )(\\d+\\.\\d+(?:\\.\\d+)?) *$\' $PACKAGE_DIR/DESCRIPTION)
[[ $VSN =~ ([0-9]+)\\.([0-9]+)(\\.([0-9]+))? ]]
MAJ=${BASH_REMATCH[1]}
MIN=${BASH_REMATCH[2]}
min=${BASH_REMATCH[4]}
if [ "$min" = "" ]; then
    min=0
fi
echo "current release number: $VSN"

read -p "Select NEW release number (current $VSN):
1) MAJOR $((MAJ+1)).0
2) MINOR $MAJ.$((MIN+1))
3) minor $MAJ.$MIN.$((min+1))
4) no change $VSN
" choice
echo 

NEWVSN=$VSN
case "$choice" in
    1) NEWVSN=$((MAJ+1)).0;;
    2) NEWVSN=$MAJ.$((MIN+1));;
    3) NEWVSN=$MAJ.$MIN.$((min+1));;
    4) NEWVSN=$NEWVSN;;
    *) echo "invalid, aborting."; exit 1;;
esac

# update version number
sed -i -r -e "s@(^Version: )($VSN) *\\$@\\1$NEWVSN@" $PACKAGE_DIR/DESCRIPTION
# TODO: hg tag?
if [ $choice -eq 1 -o $choice -eq 2 ]
then
    read -p "Add hg tag v$NEWVSN? [y/n] " addtag
    echo
    case "$addtag" in
        [yY]) echo "DO THIS:"
              echo hg tag "v$NEWVSN";;
    esac
elif [ $choice -eq 3 ]
then
    read -p "Update hg tag v$MAJ.$MIN to current revision? [y/n] " addtag
    case "$addtag" in
        [yY]) echo "DO THIS:"
              echo hg tag --remove "v$MAJ.$MIN";
              echo hg tag "v$MAJ.$MIN"
              ;;
      esac
fi     

echo "done"
',packageName),file='.precommit.sh')
            # change to executable
            Sys.chmod('.precommit.sh',mode='775')
            # add/remove
            system(sprintf('%s add',HG[1]),intern=TRUE)
            # commit
            if ( hg.commit )
                system(sprintf('%s commit -m "Initial commit"',HG[1]),intern=TRUE)
            # .hg/hgrc: only NOW add to the pre-commit hook
cat('[hooks]
pre-commit = ./.precommit.sh
',file=file.path('.hg','hgrc'))
            setwd(wd)
        }
    }

    message(sprintf("Package created, have a look at %s/README",dir))
}, error = function(e) {
    unlink(dir,recursive=TRUE)
    cat(gettext(e))
})
}


#' reload a package.
#' @param lib name of package to reload, unquoted or quoted.
#' @param force should a package be detached even though other attached
#' packages depend on it?
#' @param load boolean, whether to load the package if it wasn't already
#' loaded in the first place.
#' @param quietly boolean, suppress all warnings/messages/outputs.
#' @family library
#' @export
reload <- function(lib,force=TRUE,load=TRUE,quietly=FALSE)
{
    func <- identity
    if ( quietly )
        func <- muffle
    pkgName <- as.character(substitute(lib))
    tryCatch({
    func(detach(sprintf('package:%s',pkgName),
           character.only=TRUE,
           unload=TRUE,force=force))
    }, error=function(e)
        if ( !load )
            stop("Package ",pkgName," could not be reloaded;",
                 " perhaps it was not loaded in the first place?")
    )
    func(library(pkgName, character.only=TRUE))
}

#' muffles as much output as possible (except errors)
#' @param ... what to muffle
#' @examples
#' muffle(cat('hi'))
#' muffle(message('hi'))
#' muffle(warning('warning!'))
#' @export
#' @family output
muffle <- function(...)
{
    invisible(capture.output(
              suppressPackageStartupMessages(
              suppressMessages(
              suppressWarnings(
              ...
              )))))
}

#' muffles any output caused by warnings()/messages(), returning output
#'
#' This function suppresses any output caused by 
#' \code{\link[base]{warning}} and \code{\link[base]{message}} and 
#' \code{\link[base]{packageStartupMessage}}, and returns the result of
#' the command.
#'
#' As such, it doesn't use \code{invisible(capture.output(...))} as 
#' \code{\link{muffle}} does, and hence any output produced by 
#' \code{\link[base]{cat}} or \code{\link[base]{print}} won't be suppressed.
#'
#' @inheritParams muffle
#' @return the result of the expression in \code{...}
#' @examples
#' muffle.messages({cat('hi\n'); 1})   # returns 1, does *not* suppress 'hi'
#' muffle.messages({message('hi'); 1}) # returns 1, suppresses 'hi'
#' # returns TRUE or FALSE depending on whether you have 'tcltk' package,
#' #  and suppresses "Loading Tcl/Tk interface ... done" message, as well as
#' #  the warning (if you don't have that package).
#' muffle.messages(require(tcltk))    
#' @family output
#' @export
muffle.messages <- function(...)
{
	suppressPackageStartupMessages(
	 suppressMessages(
	  suppressWarnings(
	   ...
	  )
	 )
	)
}

#' create a string to reconstruct an object
#'
#' This captures the output of \code{\link[base]{dput}} as a string and
#' returns it.
#'
#' @param x an object
#' @return a string that is \code{dput(object)}.
#' 
#' @examples
#' stringify(c(1,2,3,4)) # 'c(1,2,3,4)'
#' @seealso \code{\link{evaluate}} to turn a stringify'd string into an R
#' object.
#' @export
stringify <- function(x)
{
    return(paste(trim(capture.output(dput(x))),collapse=''))
}

#' loads an Rdata file and returns a list of the contents
#'
#' This function loads an Rdata file and returns a list containing its
#' contents. The advantage is that it doesn't clutter the global workspace.
#'
#' @param filename name of the Rdata file
#' @param names optional, character vector of objects in the Rdata file to
#' retrieve (leave out to retrieve everything).
#' @return a LIST with \code{names(lst)==names}, values being the object
#' from the Rdata file. No entry for those objects in \code{names} that
#' were not found in the Rdata file.
#' @examples
#' # create some data to save
#' a <- 1
#' b <- 2
#' save(a,b,file='ab.rda')
#' # load them back
#' obj <- loader('ab.rda')
#' names(obj) # 'a', 'b'
#' obj$a      # '1'
#'
#' # example of looking for particular objects: look for 'a' and 'foo'.
#' obj <- loader('ab.rda',names=c('a','foo'))
#' names(obj) # 'a' (foo wasn't found)
#' obj$a      # 1
#' obj$foo    # NULL (wasn't found)
#'
#' # Demonstrate that it doesn't clutter local workspace
#' rm(list=ls())
#' a <- 'foobar'
#' obj <- loader('ab.rda',names='a')
#' obj$a  # 1
#' a      # 'foobar'
#' @export
#' @family library
loader <- function(filename,names=NULL)
{
    nms <- load(filename,(tmp<-new.env()))
    if ( is.null(names) ) names <- nms
    names <- names[names %in% nms]

    # for some reason lapply(nms,get) can't get the right environment.
    lst <- lapply( names, function(nm)
                  tryCatch(get(nm,envir=tmp), error=function(e) return(NA))
    )
    names(lst)<-names
    rm(tmp)
    return(lst)
}

